# search.py
# ---------
# Licensing Information:  You are free to use or extend these projects for 
# educational purposes provided that (1) you do not distribute or publish 
# solutions, (2) you retain this notice, and (3) you provide clear 
# attribution to UC Berkeley, including a link to 
# http://inst.eecs.berkeley.edu/~cs188/pacman/pacman.html
# 
# Attribution Information: The Pacman AI projects were developed at UC Berkeley.
# The core projects and autograders were primarily created by John DeNero 
# (denero@cs.berkeley.edu) and Dan Klein (klein@cs.berkeley.edu).
# Student side autograding was added by Brad Miller, Nick Hay, and 
# Pieter Abbeel (pabbeel@cs.berkeley.edu).


"""
In search.py, you will implement generic search algorithms which are called
by Pacman agents (in searchAgents.py).
"""

import util

class SearchProblem:
    """
    This class outlines the structure of a search problem, but doesn't implement
    any of the methods (in object-oriented terminology: an abstract class).

    You do not need to change anything in this class, ever.
    """

    def getStartState(self):
        """
        Returns the start state for the search problem
        """
        util.raiseNotDefined()

    def isGoalState(self, state):
        """
          state: Search state

        Returns True if and only if the state is a valid goal state
        """
        util.raiseNotDefined()

    def getSuccessors(self, state):
        """
          state: Search state

        For a given state, this should return a list of triples,
        (successor, action, stepCost), where 'successor' is a
        successor to the current state, 'action' is the action
        required to get there, and 'stepCost' is the incremental
        cost of expanding to that successor
        """
        util.raiseNotDefined()

    def getCostOfActions(self, actions):
        """
         actions: A list of actions to take

        This method returns the total cost of a particular sequence of actions.  The sequence must
        be composed of legal moves
        """
        util.raiseNotDefined()


def tinyMazeSearch(problem):
    """
    Returns a sequence of moves that solves tinyMaze.  For any other
    maze, the sequence of moves will be incorrect, so only use this for tinyMaze
    """
    from game import Directions
    s = Directions.SOUTH
    w = Directions.WEST
    return  [s,s,w,s,w,w,s,w]

NO_POSSIBLE_PATH = []
def getState(node):
    return node[0][0]
def getDirection(node):
    assert isinstance(node[1], basestring), node[0][1]
    return node[1]
def getExistingPath(node):
    return node[1]

def depthFirstSearch(problem):
    """
    Search the deepest nodes in the search tree first

    Your search algorithm needs to return a list of actions that reaches
    the goal.  Make sure to implement a graph search algorithm

    To get started, you might want to try some of these simple commands to
    understand the search problem that is being passed in:

    print "Start:", problem.getStartState()
    print "Is the start a goal?", problem.isGoalState(problem.getStartState())
    print "Start's successors:", problem.getSuccessors(problem.getStartState())
    """
    "*** YOUR CODE HERE ***"
    #print "Start:", problem.getStartState()
    #print "Is the start a goal?", problem.isGoalState(problem.getStartState())
    #print "Start's successors:", problem.getSuccessors(problem.getStartState())
    closed = set()
    fringe = util.Stack()
    fringe.push([[problem.getStartState(),'INIT',0], ['INIT_PATH']])
    while True:
        if fringe.isEmpty():
            return NO_POSSIBLE_PATH
        node = fringe.pop()
        if problem.isGoalState(getState(node)):
            #print "Goal is Found!: " + str(node)
            return getExistingPath(node)[1:]
        if getState(node) not in closed:
            closed.add(getState(node))
            for child_node in problem.getSuccessors(getState(node)):
                #print child_node
                child_node = [child_node, getExistingPath(node) + [getDirection(child_node)]]
                #assert len(child_node) == 2
                #print "new child: " + str(child_node)
                fringe.push(child_node)


def breadthFirstSearch(problem):
    """
    Search the shallowest nodes in the search tree first.
    """
    "*** YOUR CODE HERE ***"
    closed = set()
    fringe = util.Queue()
    fringe.push([[problem.getStartState(),'INIT',0], ['INIT_PATH']])
    while True:
        if fringe.isEmpty():
            return NO_POSSIBLE_PATH
        node = fringe.pop()
        if problem.isGoalState(getState(node)):
            #print "Goal is Found!: " + str(node)
            return getExistingPath(node)[1:]
        if getState(node) not in closed:
            closed.add(getState(node))
            for child_node in problem.getSuccessors(getState(node)):
                #print child_node
                child_node = [child_node, getExistingPath(node) + [getDirection(child_node)]]
                #assert len(child_node) == 2
                #print "new child: " + str(child_node)
                fringe.push(child_node)

def getExistingPathCost(node):
    return node[0][2]

def getStepCost(node):
    return node[2]

def uniformCostSearch(problem):
    "Search the node of least total cost first. "
    "*** YOUR CODE HERE ***"
    # http://en.wikipedia.org/wiki/Uniform-cost_search
    closed = set()
    fringe = util.PriorityQueue()
    cost = 0
    node = [[problem.getStartState(),'INIT',0], ['INIT_PATH'], cost]
    fringe.push(node, cost)
    while True:
        if fringe.isEmpty():
            return NO_POSSIBLE_PATH
        node = fringe.pop()
        if problem.isGoalState(getState(node)):
            #print "Goal is Found!: " + str(node)
            return getExistingPath(node)[1:]
        if getState(node) not in closed:
            closed.add(getState(node))
            for child_node in problem.getSuccessors(getState(node)):
                child_path = getExistingPath(node) + [getDirection(child_node)]
                cost = problem.getCostOfActions(child_path[1:])
                child_node = [child_node, child_path, cost]
                fringe.push(child_node, cost)

def nullHeuristic(state, problem=None):
    """
    A heuristic function estimates the cost from the current state to the nearest
    goal in the provided SearchProblem.  This heuristic is trivial.
    """
    return 0

def aStarSearch(problem, heuristic=nullHeuristic):
    "Search the node that has the lowest combined cost and heuristic first."
    "*** YOUR CODE HERE ***"
    closed = set()
    fringe = util.PriorityQueue()
    cost = 0
    node = [[problem.getStartState(),'INIT',0], ['INIT_PATH'], cost]
    fringe.push(node, cost)
    while True:
        if fringe.isEmpty():
            return NO_POSSIBLE_PATH
        node = fringe.pop()
        if problem.isGoalState(getState(node)):
            #print "Goal is Found!: " + str(node)
            return getExistingPath(node)[1:]
        if getState(node) not in closed:
            closed.add(getState(node))
            for child_node in problem.getSuccessors(getState(node)):
                child_path = getExistingPath(node) + [getDirection(child_node)]
                cost = problem.getCostOfActions(child_path[1:])
                child_cost = cost + heuristic(child_node[0], problem)
                child_node = [child_node, child_path, child_cost]
                fringe.push(child_node, child_cost)

# Abbreviations
bfs = breadthFirstSearch
dfs = depthFirstSearch
astar = aStarSearch
ucs = uniformCostSearch
