# multiAgents.py
# --------------
# Licensing Information:  You are free to use or extend these projects for 
# educational purposes provided that (1) you do not distribute or publish 
# solutions, (2) you retain this notice, and (3) you provide clear 
# attribution to UC Berkeley, including a link to 
# http://inst.eecs.berkeley.edu/~cs188/pacman/pacman.html
# 
# Attribution Information: The Pacman AI projects were developed at UC Berkeley.
# The core projects and autograders were primarily created by John DeNero 
# (denero@cs.berkeley.edu) and Dan Klein (klein@cs.berkeley.edu).
# Student side autograding was added by Brad Miller, Nick Hay, and 
# Pieter Abbeel (pabbeel@cs.berkeley.edu).


from util import manhattanDistance
from game import Directions
import random, util

from game import Agent

MAX, MIN = 9999999, -9999999

class ReflexAgent(Agent):
    """
      A reflex agent chooses an action at each choice point by examining
      its alternatives via a state evaluation function.

      The code below is provided as a guide.  You are welcome to change
      it in any way you see fit, so long as you don't touch our method
      headers.
    """


    def getAction(self, gameState):
        """
        You do not need to change this method, but you're welcome to.

        getAction chooses among the best options according to the evaluation function.

        Just like in the previous project, getAction takes a GameState and returns
        some Directions.X for some X in the set {North, South, West, East, Stop}
        """
        # Collect legal moves and successor states
        legalMoves = gameState.getLegalActions()

        # Choose one of the best actions
        scores = [self.evaluationFunction(gameState, action) for action in legalMoves]
        bestScore = max(scores)
        bestIndices = [index for index in range(len(scores)) if scores[index] == bestScore]
        chosenIndex = random.choice(bestIndices) # Pick randomly among the best

        "Add more of your code here if you want to"

        return legalMoves[chosenIndex]

    def evaluationFunction(self, currentGameState, action):
        """
        Design a better evaluation function here.

        The evaluation function takes in the current and proposed successor
        GameStates (pacman.py) and returns a number, where higher numbers are better.

        The code below extracts some useful information from the state, like the
        remaining food (newFood) and Pacman position after moving (newPos).
        newScaredTimes holds the number of moves that each ghost will remain
        scared because of Pacman having eaten a power pellet.

        Print out these variables to see what you're getting, then combine them
        to create a masterful evaluation function.
        """
        # Useful information you can extract from a GameState (pacman.py)
        successorGameState = currentGameState.generatePacmanSuccessor(action)
        newPos = successorGameState.getPacmanPosition()
        newFood = successorGameState.getFood().asList()
        newGhostStates = successorGameState.getGhostStates()
        newScaredTimes = [ghostState.scaredTimer for ghostState in newGhostStates]

        "*** YOUR CODE HERE ***"
        currPos = currentGameState.getPacmanPosition()
        for ghost in newGhostStates:
            ghostPos = ghost.getPosition()
            if is_near(newPos, ghostPos):
                #print "GHOST"
                return MIN
        if len(newFood) < len(currentGameState.getFood().asList()):
            #print "FOUND FOOD"
            return MAX
        curr_dist = [euclideanDistance(currPos, food) for food in newFood]
        new_dist = [euclideanDistance(newPos, food) for food in newFood]
        #from math import log10
        return 1/(sum(curr_dist) + sum(new_dist) + len(newFood)*(min(curr_dist) + min(new_dist)))

def is_near(state1, state2):
    return manhattanDistance(state1, state2) < 3
def euclideanDistance(xy1, xy2):
    return ((xy1[0]-xy2[0])**2 + (xy1[1]-xy2[1])**2)**.5

def scoreEvaluationFunction(currentGameState):
    """
      This default evaluation function just returns the score of the state.
      The score is the same one displayed in the Pacman GUI.

      This evaluation function is meant for use with adversarial search agents
      (not reflex agents).
    """
    return currentGameState.getScore()

class MultiAgentSearchAgent(Agent):
    """
      This class provides some common elements to all of your
      multi-agent searchers.  Any methods defined here will be available
      to the MinimaxPacmanAgent, AlphaBetaPacmanAgent & ExpectimaxPacmanAgent.

      You *do not* need to make any changes here, but you can if you want to
      add functionality to all your adversarial search agents.  Please do not
      remove anything, however.

      Note: this is an abstract class: one that should not be instantiated.  It's
      only partially specified, and designed to be extended.  Agent (game.py)
      is another abstract class.
    """

    def __init__(self, evalFn = 'scoreEvaluationFunction', depth = '2'):
        self.index = 0 # Pacman is always agent index 0
        self.evaluationFunction = util.lookup(evalFn, globals())
        self.depth = int(depth)

class MinimaxAgent(MultiAgentSearchAgent):
    """
      Your minimax agent (question 2)
    """

    def getAction(self, gameState):
        """
          Returns the minimax action from the current gameState using self.depth
          and self.evaluationFunction.

          Here are some method calls that might be useful when implementing minimax.

          gameState.getLegalActions(agentIndex):
            Returns a list of legal actions for an agent
            agentIndex=0 means Pacman, ghosts are >= 1

          gameState.generateSuccessor(agentIndex, action):
            Returns the successor game state after an agent takes an action

          gameState.getNumAgents():
            Returns the total number of agents in the game
        """
        "*** YOUR CODE HERE ***"
        # Lecture 6 p.16
        # http://en.wikipedia.org/wiki/Minimax#Pseudocode
        def value(self, state, action, depth, agentIndex):
            if depth == 0 or is_terminal(state, agentIndex):
                return [self.evaluationFunction(state), action]
            if agentIndex == 0:
                return max_value(self, state, action, depth, agentIndex)
            else:
                return min_value(self, state, action, depth, agentIndex)
        def max_value(self, state, from_action, depth, agentIndex):
            assert agentIndex == 0, agentIndex
            v = (MIN, "MIN_ACTION")
            actions = state.getLegalActions(agentIndex)
            for action in actions:
                successor = value(self, state.generateSuccessor(agentIndex, action), action, depth, get_next_agent(agentIndex))
                if utility(v) < successor:
                    v = [utility(successor), action]
            assert v != (MIN, "MIN_ACTION")
            return v
        def min_value(self, state, from_action, depth, agentIndex):
            assert agentIndex > 0, agentIndex
            v = (MAX, "MAX_ACTION")
            actions = state.getLegalActions(agentIndex)
            if agentIndex == gameState.getNumAgents() - 1:
                depth -= 1
            for action in actions:
                successor = value(self, state.generateSuccessor(agentIndex, action), action, depth, get_next_agent(agentIndex))
                if utility(v) > utility(successor):
                    v = [utility(successor), action]
            assert v != (MAX, "MAX_ACTION")
            return v
        def is_terminal(state, agentIndex):
            return not state.getLegalActions(agentIndex)
        def utility(state):
            assert isinstance(state[0], float) or isinstance(state[0], int), state
            return state[0]
        def get_action(state):
            assert type(state[1]) == basestring
            return state[1]
        def get_next_agent(agentIndex):
            return (agentIndex + 1) % gameState.getNumAgents()
        minimax = value(self, gameState, "ACTION", self.depth, 0)
        print minimax
        return minimax[1]



class AlphaBetaAgent(MultiAgentSearchAgent):
    """
      Your minimax agent with alpha-beta pruning (question 3)
    """

    def getAction(self, gameState):
        """
          Returns the minimax action using self.depth and self.evaluationFunction
        """
        "*** YOUR CODE HERE ***"
        util.raiseNotDefined()

class ExpectimaxAgent(MultiAgentSearchAgent):
    """
      Your expectimax agent (question 4)
    """

    def getAction(self, gameState):
        """
          Returns the expectimax action using self.depth and self.evaluationFunction

          All ghosts should be modeled as choosing uniformly at random from their
          legal moves.
        """
        "*** YOUR CODE HERE ***"
        util.raiseNotDefined()

def betterEvaluationFunction(currentGameState):
    """
      Your extreme ghost-hunting, pellet-nabbing, food-gobbling, unstoppable
      evaluation function (question 5).

      DESCRIPTION: <write something here so we know what you did>
    """
    "*** YOUR CODE HERE ***"
    util.raiseNotDefined()

# Abbreviation
better = betterEvaluationFunction

